/* 
 * Support for CE-InfoSys FastCrypt hardware crypto engine.
 *
 * Linux developers:
 * 	Michal Ludvig <michal@logix.cz>
 * 	              http://www.logix.cz/michal
 *
 * Details about communication with the card learned from fcrypt-2.00
 * driver written by Benjamin Reichardt <reichard@math.uni-goettingen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>

#include <linux/crypto.h>

#include <linux/delay.h>
#include <linux/wait.h>

#include <linux/interrupt.h>
#include <linux/pci.h>

#include "fcrypt.h"
#include "fcrypt_refdat.h"

#define	FCRYPT_MODULE_NAME "fcrypt"
#define	PFX FCRYPT_MODULE_NAME ": "

MODULE_AUTHOR("Michal Ludvig <michal@logix.cz>");
MODULE_DESCRIPTION("CE-Infosys FastCrypt PCI crypto accelerator driver");
MODULE_LICENSE("GPL");

#ifdef CONFIG_SUSE_KERNEL
MODULE_INFO(supported, "yes");
#endif

static struct pci_device_id fcrypt_pci_tbl[] =
{
	{0x10e8, 0x80ad, PCI_ANY_ID, PCI_ANY_ID, 0, 0, },
	{ }
};
MODULE_DEVICE_TABLE(pci, fcrypt_pci_tbl);

struct fcrypt_list flist_global;

/* ====== Functions for accessing indirect ports ====== */

static inline void
fcrypt_aport_iwrite_v(struct fcrypt_dev *fdev,
		      unsigned long iregister,
		      int count, ...)
{
	va_list ap;
	
	outl(A_PORT_APTR_OFFSET, fdev->adrctr);
	outl(iregister, fdev->aport);
	outl(A_PORT_AIAR_OFFSET, fdev->adrctr);
	
	va_start(ap, count);
	while(count--)
		outl(va_arg(ap, unsigned long), fdev->aport);
	va_end(ap);
}

static inline void
fcrypt_aport_iwrite(struct fcrypt_dev *fdev,
		    unsigned int iregister,
		    unsigned long value)
{
	outl(A_PORT_APTR_OFFSET, fdev->adrctr);
	outl(iregister, fdev->aport);
	outl(A_PORT_AIAR_OFFSET, fdev->adrctr);
	outl(value, fdev->aport);
}

static inline void 
fcrypt_cport_iwrite_v(struct fcrypt_dev *fdev,
		      unsigned long iregister,
		      int count, ...)
{
	va_list ap;
	
	outl(C_PORT_CPTR_OFFSET, fdev->adrctr);
	outl(iregister, fdev->cport);
	outl(C_PORT_CIAR_OFFSET, fdev->adrctr);
	
	va_start(ap, count);
	while(count--)
		outl(va_arg(ap, unsigned long), fdev->cport);
	va_end(ap);
}

static inline void
fcrypt_cport_iwrite(struct fcrypt_dev *fdev,
		    unsigned int iregister,
		    unsigned long value)
{
	outl(C_PORT_CPTR_OFFSET, fdev->adrctr);
	outl(iregister, fdev->cport);
	outl(C_PORT_CIAR_OFFSET, fdev->adrctr);
	outl(value, fdev->cport);
}

/* ====== Functions for managing fcrypt_list ====== */

static void 
fcrypt_add_device(struct fcrypt_dev *fdev)
{
	spin_lock(&flist_global.lock);
	list_add(&fdev->entry, &flist_global.list);
	flist_global.count++;
	spin_unlock(&flist_global.lock);
}

static void 
fcrypt_remove_device(struct fcrypt_dev *fdev)
{
	struct fcrypt_dev *fdev_ptr, *tmp;

	struct list_head *head = &flist_global.list;

	spin_lock(&flist_global.lock);
	list_for_each_entry_safe(fdev_ptr, tmp, head, entry)
		if (fdev_ptr == fdev) {
			list_del(&fdev_ptr->entry);
			flist_global.count--;
			break;
		}
	spin_unlock(&flist_global.lock);
}

static struct fcrypt_dev *
fcrypt_get_device(void)
{
	struct fcrypt_dev *fdev_ptr;
	
	spin_lock(&flist_global.lock);
	list_for_each_entry(fdev_ptr, &flist_global.list, entry)
		if (fdev_ptr->status == BOARD_FREE)
			break;
	spin_unlock(&flist_global.lock);

	return fdev_ptr;
}

static void
fcrypt_put_device(struct fcrypt_dev *fdev)
{
}

/* ====== Functions for loading keys and IVs into the card ====== */

static void
fcrypt_load_one_key(struct fcrypt_dev *fdev, const char *key, int slot)
{
	int i;
	
	BUG_ON(slot & ~0x0F);

	fcrypt_cport_iwrite(fdev, C_PORT_CMD1_IOFFSET, slot << 4);
	fcrypt_cport_iwrite(fdev, C_PORT_CMD2_IOFFSET, 0);

	fcrypt_cport_iwrite_v(fdev, C_PORT_CDP_IOFFSET, 0);
	
	for (i = 0; i < 8; i++)
		outl(key[i], fdev->cport);

	outl(C_PORT_CSTAT_OFFSET, fdev->adrctr);
}

static void
fcrypt_load_key(struct fcrypt_dev *fdev, const char *key,
		int key_slots[3], int mode)
{
	if (mode & TRIPLE_CIPHER) {
		if ((mode & ENCDEC_MASK) == ENCODE) {
			fcrypt_load_one_key(fdev, key +  0, key_slots[0]);
			fcrypt_load_one_key(fdev, key +  8, key_slots[1]);
			fcrypt_load_one_key(fdev, key + 16, key_slots[2]);
		} else {
			/* 3DES decryption needs the key in reversed order. */
			fcrypt_load_one_key(fdev, key + 16, key_slots[0]);
			fcrypt_load_one_key(fdev, key +  8, key_slots[1]);
			fcrypt_load_one_key(fdev, key +  0, key_slots[2]);
		}
	} else
		fcrypt_load_one_key(fdev, key, key_slots[0]);
}

static void
fcrypt_load_iv(struct fcrypt_dev *fdev, const char *iv)
{
	int i;
	
	fcrypt_aport_iwrite_v(fdev, A_PORT_IV_IOFFSET, 0);
	
	for (i = 0; i < 8; i++)
		outl(iv[i], fdev->aport);

	fcrypt_aport_iwrite(fdev, A_PORT_CMD_IOFFSET, A_PORT_StopStoreIV1_CMD);
}

/* ====== Helper function ====== */

static void 
dump_mem(const char *msg, const char *ibuf, size_t len)
{
	char obuf[100];
	int i = 0;

	memset(obuf, 0, sizeof(obuf));
	for (i=0; i<len; i++) {
		sprintf(&obuf[(i%8)*2], "%02x", ibuf[i] & 0xFF);
		if ((i % 8 == 7) || (i==len - 1))
			printk(KERN_INFO PFX "dump: %s %s\n", msg, obuf);
	}
}

/* ====== Encryption/decryption functions ====== */

static void
fcrypt_crypt_dma(struct fcrypt_dev *fdev, int len, int key_slot[3], int mode)
{
	/* Set board status */
	fdev->status = BOARD_BUSY;

	/* Reset PCI FIFO */
	outl(PCI_RESET_CMD, (fdev->opreg + MCSR_REG_OFFSET));

	/* Reset A-Port */
	fcrypt_aport_iwrite(fdev, A_PORT_CMD_IOFFSET, A_PORT_Reset_CMD);

	/* Enable B-Port */
	fcrypt_aport_iwrite(fdev, A_PORT_CFG_IOFFSET, CFG_ENABLE_B_RW);

	/* Enable DMA */
	fcrypt_aport_iwrite(fdev, A_PORT_DMA_IOFFSET, DMA_ENABLE);

	/* Set used key slots */
	fcrypt_aport_iwrite(fdev, A_PORT_KPTR_IOFFSET, key_slot[0]);
	if (mode & TRIPLE_CIPHER) {
		fcrypt_aport_iwrite(fdev, A_PORT_KPTR_IOFFSET + 1, key_slot[1]);
		fcrypt_aport_iwrite(fdev, A_PORT_KPTR_IOFFSET + 2, key_slot[2]);
	}

	/* Set chaining mode */
	fcrypt_aport_iwrite(fdev, A_PORT_MOD_IOFFSET, mode);

	/* Initiate BusMaster transfer */
	/* Input buffer */
	outl(fdev->ipage_phys, fdev->opreg + MRAR_REG_OFFSET);
	outl(len, fdev->opreg + MRTC_REG_OFFSET);
	
	/* Output buffer */
	outl(fdev->opage_phys, fdev->opreg + MWAR_REG_OFFSET);
	outl(len, fdev->opreg + MWTC_REG_OFFSET);

	/* Enable IRQ and BusMaster transfers in both directions */
	outl(INT_ON_WRITE_COMPLETE, fdev->opreg + INTR_CTRL_REG_OFFSET);
	outl(PCI_BUSMASTER_WRITE_ENABLE + PCI_BUSMASTER_READ_ENABLE,
	     fdev->opreg + MCSR_REG_OFFSET);

	/* Start encryption with IV#1 */
	fcrypt_aport_iwrite(fdev, A_PORT_CMD_IOFFSET, A_PORT_Fetch1Start_CMD);

	/* Enable B-Port transfer */
	outl(AdrCtl_EBT_BIT + A_PORT_ADATA_OFFSET, fdev->adrctr);
}

static int
fcrypt_run_sync(struct fcrypt_dev *fdev, const uint8_t *iv,
		int len, int key_slots[3], int mode)
{
	int rc = 0;
	wait_queue_head_t waitq;
	unsigned long timeout;

	if (iv)
		fcrypt_load_iv(fdev, iv);

	if (!in_atomic()) {
		fdev->pwaitq = &waitq;
		init_waitqueue_head(fdev->pwaitq);
	} else
		fdev->pwaitq = NULL;
	
	fcrypt_crypt_dma(fdev, len, key_slots, mode);

	if (!in_atomic())
		wait_event_interruptible_timeout(*fdev->pwaitq, 
						 (fdev->status != BOARD_BUSY),
						 2 * HZ);
	else {
		timeout = jiffies + 2 * HZ;
		while((fdev->status == BOARD_BUSY) && jiffies < timeout)
			;
		if (fdev->status == BOARD_BUSY)
			printk(KERN_CRIT PFX "%s(): Timeout!!!\n", __func__);
	}

	if (unlikely(fdev->status != BOARD_CRYPT_OK)) {
		printk(KERN_CRIT PFX "Reported bad status! (status=%d)\n", fdev->status);
		fdev->status = BOARD_FAILED;
	} else
		fdev->status = BOARD_FREE;

	return rc;
}

static int
fcrypt_3des_set_key(void *ctx_arg, const uint8_t *in_key,
		    unsigned int key_len, uint32_t *flags)
{
	struct fcrypt_3des_ctx *ctx = ctx_arg;

	if (key_len != DES3_KEY_SIZE) {
		*flags |= CRYPTO_TFM_RES_BAD_KEY_LEN;
		return -EINVAL;
	}

	memcpy(ctx->key, in_key, key_len);

	return 0;
}

static void
fcrypt_3des_singleblock(void *ctx_arg, uint8_t *dst, const uint8_t *src, int encdec)
{
	struct fcrypt_dev *fdev;
	int key_slots[3] = { 1, 2, 3 };
	struct fcrypt_3des_ctx *ctx = ctx_arg;

	fdev = fcrypt_get_device();
	fcrypt_load_key(fdev, ctx->key, key_slots, TRIPLE_CIPHER | encdec);

	memcpy(fdev->ipage, src, DES3_BLOCK_SIZE);
	fcrypt_run_sync(fdev, NULL, DES3_BLOCK_SIZE, key_slots, ECB_MODE | TRIPLE_CIPHER | encdec);
	memcpy(dst, fdev->opage, DES3_BLOCK_SIZE);

	fcrypt_put_device(fdev);
}

static void
fcrypt_3des_encrypt(void *ctx_arg, uint8_t *dst, const uint8_t *src)
{
	fcrypt_3des_singleblock(ctx_arg, dst, src, ENCODE);
}

static void
fcrypt_3des_decrypt(void *ctx_arg, uint8_t *dst, const uint8_t *src)
{
	fcrypt_3des_singleblock(ctx_arg, dst, src, DECODE);
}

static void
fcrypt_3des_multiblock(void *ctx_arg, uint8_t *dst, const uint8_t *src,
		       const uint8_t *iv, size_t nbytes, int encdec,
		       int mode)
{
	struct fcrypt_dev *fdev = fcrypt_get_device();
	int key_slots[3] = { 1, 2, 3 };
	struct fcrypt_3des_ctx *ctx = ctx_arg;
	size_t chunk, index = 0, todo = nbytes;

	fcrypt_load_key(fdev, ctx->key, key_slots, TRIPLE_CIPHER | encdec);

	while (todo) {
		chunk = todo > PAGE_SIZE ? PAGE_SIZE : todo;
		
		memcpy(fdev->ipage, src + index, chunk);
		fcrypt_run_sync(fdev, iv, chunk, key_slots, TRIPLE_CIPHER | mode | encdec);
		memcpy(dst + index, fdev->opage, chunk);

		index += chunk;
		todo -= chunk;
		iv = NULL;	/* Actual IV is already stored on the chip */
	}
	fcrypt_put_device(fdev);
}

static void
fcrypt_3des_ecb(void *ctx_arg, uint8_t *dst, const uint8_t *src,
		const uint8_t *iv, size_t nbytes, int encdec,
		int inplace)
{
	fcrypt_3des_multiblock(ctx_arg, dst, src, iv, nbytes,
			       encdec == CRYPTO_DIR_ENCRYPT ? ENCODE : DECODE,
			       ECB_MODE);
}

static void
fcrypt_3des_cbc(void *ctx_arg, uint8_t *dst, const uint8_t *src,
		const uint8_t *iv, size_t nbytes, int encdec,
		int inplace)
{
	fcrypt_3des_multiblock(ctx_arg, dst, src, iv, nbytes,
			       encdec == CRYPTO_DIR_ENCRYPT ? ENCODE : DECODE,
			       CBC_MODE);
}

static void
fcrypt_3des_cfb(void *ctx_arg, uint8_t *dst, const uint8_t *src,
		const uint8_t *iv, size_t nbytes, int encdec,
		int inplace)
{
	fcrypt_3des_multiblock(ctx_arg, dst, src, iv, nbytes,
			       encdec == CRYPTO_DIR_ENCRYPT ? ENCODE : DECODE,
			       CFB_MODE);
}

static void
fcrypt_3des_ofb(void *ctx_arg, uint8_t *dst, const uint8_t *src,
		const uint8_t *iv, size_t nbytes, int encdec,
		int inplace)
{
	fcrypt_3des_multiblock(ctx_arg, dst, src, iv, nbytes,
			       encdec == CRYPTO_DIR_ENCRYPT ? ENCODE : DECODE,
			       OFB_MODE);
}

static int
fcrypt_test(struct fcrypt_dev *fdev)
{
	int key_slots[3] = { 1, 2, 3 };
	int i, rc = 0;

	for (i = 0; i < num_fcrypt_test_vectors; i++) {
		memcpy(fdev->ipage, fcrypt_test_vectors[i].ref_input,
		       fcrypt_test_vectors[i].bytes);

		fcrypt_load_key(fdev, fcrypt_test_vectors[i].key,
				key_slots, fcrypt_test_vectors[i].mode);

		fcrypt_run_sync(fdev, fcrypt_test_vectors[i].iv,
				fcrypt_test_vectors[i].bytes,
				key_slots,
				fcrypt_test_vectors[i].mode);

		if (memcmp(fdev->opage, fcrypt_test_vectors[i].ref_output,
			   fcrypt_test_vectors[i].bytes) != 0) {
			printk(KERN_ERR PFX "%s FAILED\n", fcrypt_test_vectors[i].label);
			dump_mem("Input:    ", fdev->ipage, fcrypt_test_vectors[i].bytes);
			dump_mem("Expected: ", fcrypt_test_vectors[i].ref_output,
				 fcrypt_test_vectors[i].bytes);
			dump_mem("Got:      ", fdev->opage, fcrypt_test_vectors[i].bytes);

			rc = -ENODEV;

			break;
		}
	}

	return rc;
}

static irqreturn_t
fcrypt_interrupt(int irq, void *dev_instance, struct pt_regs *rgs)
{
	struct fcrypt_dev *fdev = dev_instance;
	unsigned long status;

	status = inl(fdev->opreg + INTR_CTRL_REG_OFFSET);

	/* Request from our card? */
	if ((status & INT_ASSERTED) == 0)
		return IRQ_NONE;

	if ((status & (WRITE_TRANSFER_COMPLETE | READ_TRANSFER_COMPLETE))) {
		/* Disable BusMaster */
		outl(0x00000000, fdev->opreg + MCSR_REG_OFFSET);

		/* Disable IRQ */
		outl(0x003f0000, fdev->opreg + INTR_CTRL_REG_OFFSET);

		/* Reset EBT */
		outl(0, fdev->adrctr);

		/* Stop CryptoEngine and store IV to slot #1 */
		fcrypt_aport_iwrite(fdev, A_PORT_CMD_IOFFSET, A_PORT_StopStoreIV1_CMD);

		/* Set board status */
		fdev->status = BOARD_CRYPT_OK;
	} else {
		/* Set board status */
		fdev->status = BOARD_WRONG_INT;
		
		printk(KERN_CRIT PFX "received undefined IRQ\n");
	}

	if (fdev->status_report)
		*fdev->status_report = fdev->status;

	if (fdev->pwaitq)
		wake_up(fdev->pwaitq);

	return IRQ_HANDLED;
}

/* ====== Card initialization routines ====== */

static int
fcrypt_init_hardware(struct fcrypt_dev *fdev)
{
	int i;
	unsigned char xbox_read;

	/* Reset the PCI chip */
	outl(PCI_RESET_CMD, (unsigned long) (fdev->opreg + MCSR_REG_OFFSET));

	/* Reset the SuperCrypt chip */
	outl(PCI_FULL_RESET_CMD, (unsigned int)(fdev->opreg + MCSR_REG_OFFSET));

	/* Clear MSCR */
	outl(0, (unsigned int)(fdev->opreg + MCSR_REG_OFFSET));

	/* Reset C-Port */
	fcrypt_cport_iwrite(fdev, C_PORT_CMD1_IOFFSET, C_PORT_Reset_CMD1);

	/* Clear C-Port */
	fcrypt_cport_iwrite(fdev, C_PORT_CMD1_IOFFSET, 0);

	/* Load X-Box */
	fcrypt_cport_iwrite(fdev, C_PORT_CMD2_IOFFSET, C_PORT_XBoxLoad_CMD2);
	
	fcrypt_cport_iwrite_v(fdev, C_PORT_CDP_IOFFSET, 0);

	for (i = 0; i < X_BOX_SIZE; i++)
		outl(gXBox_Data[i], fdev->cport);

	for (i = 0; i < X_BOX_SIZE; i++) {
		xbox_read = inb(fdev->cport);
		if (xbox_read != gXBox_Data[i]) {
			printk(KERN_CRIT PFX "X-Box verification failed (%x/%x@%d)!\n",
			       xbox_read & 0xff, gXBox_Data[i] & 0xff, i);
			return -EACCES;
		}
	}

	/* Reset C-Port */
	fcrypt_cport_iwrite(fdev, C_PORT_CMD1_IOFFSET, C_PORT_Reset_CMD1);

	/* Enable all key slots */
	fcrypt_cport_iwrite_v(fdev, C_PORT_KYE_IOFFSET, 2, 0x000000ff, 0x000000ff);

	/* Enable A-Port */
	fcrypt_cport_iwrite(fdev, C_PORT_CFG_IOFFSET,
			    CFG_ENABLE_A_PORT + CFG_ENABLE_A_PORT_IV);

	return 0;
}

static int __devinit
fcrypt_init_one(struct pci_dev *pdev, const struct pci_device_id *ent)
{
	int		rc;
	struct fcrypt_dev *fdev;

	rc = pci_enable_device(pdev);
	if (rc) {
		printk(KERN_ERR PFX "Cannot enable PCI card. Aborting.\n");
		return rc;
	}

	rc = pci_set_dma_mask(pdev, 0xffffffff);
	if (rc) {
		printk(KERN_ERR PFX "DMA not supported. Aborting.\n");
		goto err_out_disable;
	}

	rc = pci_request_regions(pdev, FCRYPT_MODULE_NAME);
	if (rc) {
		printk(KERN_ERR PFX "Cannot register resources. Aborting.\n");
		goto err_out_disable;
	}

	pci_set_master(pdev);

	fdev = kmalloc(sizeof(struct fcrypt_dev), GFP_KERNEL);
	if (!fdev) {
		printk(KERN_ERR PFX "Memory allocation failed. Aborting.\n");
		rc = -ENOMEM;
		goto err_out_relreg;
	}

	memset(fdev, 0, sizeof(struct fcrypt_dev));

	fdev->pdev   = pdev;
	fdev->opreg  = pci_resource_start(pdev, 0);
	fdev->adrctr = pci_resource_start(pdev, 1);
	fdev->aport  = pci_resource_start(pdev, 2);
	fdev->cport  = pci_resource_start(pdev, 3);

	// spin_lock_init(&fdev->lock);
	fdev->status = BOARD_FREE;

	fdev->ipage  = (void*)get_zeroed_page(GFP_DMA);
        fdev->opage  = (void*)get_zeroed_page(GFP_DMA);
	if (!fdev->ipage || !fdev->opage) {
		printk(KERN_ERR PFX "Unable to get free page.\n");
		rc = -ENOMEM;
		goto err_out_freemem;
	}
	fdev->ipage_phys = virt_to_bus(fdev->ipage);
	fdev->opage_phys = virt_to_bus(fdev->opage);

	pci_set_drvdata(pdev, fdev);

	rc = request_irq(pdev->irq, fcrypt_interrupt,
			 SA_INTERRUPT | SA_SHIRQ, "fcrypt", fdev);
	if (rc) {
		printk(KERN_ERR PFX "Cannot register for IRQ %d\n", pdev->irq);
		goto err_out_freemem;
	}

	rc = fcrypt_init_hardware(fdev);
	if (rc) {
		printk(KERN_ERR PFX "Hardware initialization failed. Aborting.\n");
		goto err_out_freeirq;
	}

	rc = fcrypt_test(fdev);
	if (rc) {
		printk(KERN_ERR PFX "Hardware verification failed. Aborting.\n");
		goto err_out_freeirq;
	}

	fcrypt_add_device(fdev);

#if 0
	printk(KERN_INFO PFX "driver configured for 0x%lx/0x%lx/0x%lx/0x%lx, IRQ=%d\n",
	       fdev->opreg, fdev->adrctr, fdev->aport, fdev->cport, fdev->pdev->irq);
#endif
	printk(KERN_INFO PFX "driver configured for %s, IRQ=%d\n", fdev->pdev->slot_name, 
	       fdev->pdev->irq);

	return 0;

err_out_freeirq:
	free_irq(pdev->irq, fdev);

err_out_freemem:
	if (fdev) {
		if (fdev->ipage) 
			free_page((unsigned long)(fdev->ipage));

		if (fdev->opage)
			free_page((unsigned long)(fdev->opage));

		kfree(fdev);
	}

err_out_relreg:
	pci_release_regions(pdev);

err_out_disable:
	pci_disable_device(pdev);
	pci_set_drvdata(pdev, NULL);

	return rc;
}

static void __devexit
fcrypt_remove_one(struct pci_dev *pdev)
{
	struct fcrypt_dev *fdev = pci_get_drvdata(pdev);

	free_irq(pdev->irq, fdev);

	pci_release_regions(pdev);
	pci_set_drvdata(pdev, NULL);
	pci_disable_device(pdev);

	fcrypt_remove_device(fdev);

	if (fdev->ipage) 
		free_page((unsigned long)(fdev->ipage));

	if (fdev->opage)
		free_page((unsigned long)(fdev->opage));

	if (fdev)
		kfree(fdev);
}

/* ====== Module services ====== */

static struct pci_driver fcrypt_driver = {
	.name		= FCRYPT_MODULE_NAME,
	.id_table	= fcrypt_pci_tbl,
	.probe		= fcrypt_init_one,
	.remove		= __devexit_p(fcrypt_remove_one),
};

static struct crypto_alg fcrypt_3des_alg = {
	.cra_name		= "des3_ede",
	.cra_preference		= CRYPTO_PREF_HARDWARE,
	.cra_flags		= CRYPTO_ALG_TYPE_CIPHER,
	.cra_blocksize		= DES3_BLOCK_SIZE,
	.cra_ctxsize		= sizeof(struct fcrypt_3des_ctx),
	.cra_module		= THIS_MODULE,
	.cra_list		= LIST_HEAD_INIT(fcrypt_3des_alg.cra_list),
	.cra_u			= {
		.cipher = {
			.cia_min_keysize	= DES3_KEY_SIZE,
			.cia_max_keysize	= DES3_KEY_SIZE,
			.cia_setkey	   	= fcrypt_3des_set_key,
			.cia_encrypt	 	= fcrypt_3des_encrypt,
			.cia_decrypt	  	= fcrypt_3des_decrypt,

			.cia_max_nbytes		= (size_t)-1,
			.cia_ecb		= fcrypt_3des_ecb,
			.cia_cbc		= fcrypt_3des_cbc,
			.cia_cfb		= fcrypt_3des_cfb,
			.cia_ofb		= fcrypt_3des_ofb,
		}
	}
};

static int __init
fcrypt_init(void)
{
	int rc;

	INIT_LIST_HEAD(&flist_global.list);
	spin_lock_init(&flist_global.lock);

	rc = pci_module_init(&fcrypt_driver);
	if (rc)
		return rc;

	rc = crypto_register_alg(&fcrypt_3des_alg);
	if (rc) {
		pci_unregister_driver(&fcrypt_driver);
		return rc;
	}

	printk(KERN_INFO PFX "FastCrypt driver found %d board%s\n",
	       flist_global.count, flist_global.count > 1 ? "s" : "");

	return 0;
}

static void __exit
fcrypt_exit(void)
{
	crypto_unregister_alg(&fcrypt_3des_alg);
	pci_unregister_driver(&fcrypt_driver);

	printk(KERN_INFO PFX "driver unloaded.\n");
}

module_init(fcrypt_init);
module_exit(fcrypt_exit);
